# -*- coding: utf-8 -*-
#
# Copyright (C) 2018 CERN.
# Copyright (C) 2018 RERO.
#
# Invenio-Circulation is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Pytest configuration."""

from __future__ import absolute_import, print_function

import json
import uuid
from os.path import dirname, join

import mock
import pytest
from invenio_access import ActionRoles, superuser_access
from invenio_accounts.models import Role
from invenio_app.factory import create_api
from invenio_db import db
from invenio_indexer.api import RecordIndexer
from invenio_records_rest.utils import allow_all
from invenio_search import current_search

from invenio_circulation.api import Loan
from invenio_circulation.permissions import loan_access
from invenio_circulation.pidstore.minters import loan_pid_minter

from .helpers import create_loan, test_views_permissions_factory


@pytest.fixture(scope='module')
def create_app():
    """Return API app."""
    return create_api


@pytest.fixture(scope="module")
def app_config(app_config):
    """Flask application fixture."""
    app_config["JSONSCHEMAS_ENDPOINT"] = "/schema"
    app_config["JSONSCHEMAS_HOST"] = "localhost:5000"
    app_config["RECORDS_REST_DEFAULT_READ_PERMISSION_FACTORY"] = allow_all
    app_config["CIRCULATION_ITEM_EXISTS"] = lambda x: True
    app_config["CIRCULATION_PATRON_EXISTS"] = lambda x: True
    return app_config


@pytest.fixture()
def loan_created(app):
    """Minimal Loan object."""
    record_uuid = uuid.uuid4()
    new_loan = {}
    loan_pid_minter(record_uuid, data=new_loan)
    loan = Loan.create(data=new_loan, id_=record_uuid)
    db.session.commit()
    yield loan


@pytest.fixture()
def params():
    """Params for API REST payload."""
    return dict(
        transaction_user_pid="user_pid",
        patron_pid="patron_pid",
        item_pid="item_pid",
        transaction_location_pid="loc_pid",
        transaction_date="2018-02-01T09:30:00+02:00",
    )


@pytest.fixture()
def json_headers(app):
    """JSON headers."""
    return [
        ("Content-Type", "application/json"),
        ("Accept", "application/json"),
    ]


@pytest.fixture(scope="session")
def test_data():
    """Load test records."""
    path = "data/loans.json"
    with open(join(dirname(__file__), path)) as fp:
        loans = json.load(fp)
    yield loans


@pytest.fixture()
def test_loans(db, test_data):
    """Load test records."""
    loans = []
    for data in test_data:
        loans.append(create_loan(data))
    db.session.commit()
    yield loans


@pytest.fixture()
def indexed_loans(es, test_loans):
    """Get a function to wait for records to be flushed to index."""
    indexer = RecordIndexer()
    for pid, loan in test_loans:
        indexer.index(loan)
    current_search.flush_and_refresh(index="loans")

    yield test_loans

    for pid, loan in test_loans:
        indexer.delete_by_id(loan.id)
    current_search.flush_and_refresh(index="loans")


@pytest.fixture()
def mock_is_item_available():
    """Mock item_available check."""
    path = "invenio_circulation.transitions.base.is_item_available"
    with mock.patch(path) as mock_is_item_available:
        mock_is_item_available.return_value = True
        yield mock_is_item_available


@pytest.fixture()
def users(db, base_app):
    """Create admin, manager and user."""
    base_app.config["CIRCULATION_VIEWS_PERMISSIONS_FACTORY"] = \
        test_views_permissions_factory

    with db.session.begin_nested():
        datastore = base_app.extensions['security'].datastore

        # create users
        manager = datastore.create_user(email='manager@test.com',
                                        password='123456', active=True)
        admin = datastore.create_user(email='admin@test.com',
                                      password='123456', active=True)
        user = datastore.create_user(email='user@test.com',
                                     password='123456', active=True)

        # Give role to admin
        admin_role = Role(name='admin')
        db.session.add(ActionRoles(action=superuser_access.value,
                                   role=admin_role))
        datastore.add_role_to_user(admin, admin_role)
        # Give role to user
        manager_role = Role(name='manager')
        db.session.add(ActionRoles(action=loan_access.value,
                                   role=manager_role))
        datastore.add_role_to_user(manager, manager_role)
    db.session.commit()

    return {
        'admin': admin,
        'manager': manager,
        'user': user,
    }



# -*- coding: utf-8 -*-
#
# Copyright (C) 2018 CERN.
# Copyright (C) 2018 RERO.
#
# Invenio-Circulation is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Circulation minters."""

from ..api import Loan
from .providers import CirculationLoanIdProvider


def loan_pid_minter(record_uuid, data):
    """Mint loan identifiers."""
    assert Loan.pid_field not in data
    provider = CirculationLoanIdProvider.create(
        object_type='rec',
        object_uuid=record_uuid,
    )
    data[Loan.pid_field] = provider.pid.pid_value
    return provider.pid



# -*- coding: utf-8 -*-
#
# Copyright (C) 2018 CERN.
# Copyright (C) 2018 RERO.
#
# Invenio-Circulation is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Tests for loan search class."""

from invenio_circulation.api import Loan
from invenio_circulation.search.api import search_by_patron_item, \
    search_by_patron_pid, search_by_pid


def test_search_loans_by_pid(indexed_loans):
    """Test retrieve loan list belonging to an item."""
    loans = list(search_by_pid(item_pid="item_pending_1").scan())
    assert len(loans) == 1
    loan = Loan.get_record_by_pid(loans[0][Loan.pid_field])
    assert loan.get("item_pid") == "item_pending_1"


def test_search_loans_by_pid_filtering_states(indexed_loans):
    """Test retrieve loan list belonging to an item filtering states."""
    search = search_by_pid(item_pid="item_multiple_pending_on_loan_7",
                           filter_states=["PENDING", "ITEM_ON_LOAN"])
    search_result = search.execute()
    assert search_result.hits.total == 3


def test_search_loans_by_pid_excluding_states(indexed_loans):
    """Test retrieve loan list belonging to an item excluding states."""
    search_result = search_by_pid(item_pid="item_multiple_pending_on_loan_7",
                                  exclude_states=["ITEM_ON_LOAN"]).execute()
    assert search_result.hits.total == 2


def test_search_loans_by_patron_pid(indexed_loans):
    """Test retrieve loan list belonging to a patron."""
    search_result = search_by_patron_pid("1").execute()
    assert search_result.hits.total == 8

    search_result = search_by_patron_pid("2").execute()
    assert search_result.hits.total == 3

    search_result = search_by_patron_pid("3").execute()
    assert search_result.hits.total == 1


def test_search_loans_by_patron_and_item(indexed_loans):
    """Test retrieve loan list by patron and items."""
    search_result = search_by_patron_item(patron_pid="1",
                                          item_pid="item_returned_3").execute()
    assert search_result.hits.total == 1

    search_result = search_by_patron_item(patron_pid="1",
                                          item_pid="not_existing").execute()
    assert search_result.hits.total == 0

    search_result = search_by_patron_item(patron_pid="999999",
                                          item_pid="item_returned_3").execute()
    assert search_result.hits.total == 0


def test_search_loans_by_patron_and_item_filtering_states(indexed_loans):
    """Test retrieve loan list by patron and items filtering states."""
    search = search_by_patron_item(patron_pid="1",
                                   item_pid="item_returned_3",
                                   filter_states=['ITEM_RETURNED'])
    search_result = search.execute()
    assert search_result.hits.total == 1

    search = search_by_patron_item(patron_pid="1",
                                   item_pid="item_returned_3",
                                   filter_states=['ITEM_AT_DESK'])
    search_result = search.execute()
    assert search_result.hits.total == 0



# -*- coding: utf-8 -*-
#
# Copyright (C) 2018 CERN.
# Copyright (C) 2018 RERO.
#
# Invenio-Circulation is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Circulation Item JSON Resolver module."""

import jsonresolver
from werkzeug.routing import Rule


@jsonresolver.hookimpl
def jsonresolver_loader(url_map):
    """Resolve the OpenAIRE grant."""
    from flask import current_app
    resolving_path = current_app.config.get(
        "CIRCULATION_ITEM_RESOLVING_PATH") or "/"
    url_map.add(Rule(
        resolving_path,
        endpoint=current_app.config.get('CIRCULATION_ITEM_RESOLVER_ENDPOINT'),
        host=current_app.config.get('JSONSCHEMAS_HOST')))



# -*- coding: utf-8 -*-
#
# Copyright (C) 2018 CERN.
# Copyright (C) 2018 RERO.
#
# Invenio-Circulation is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Module tests."""

from __future__ import absolute_import, print_function

import json

from flask import url_for


def test_version():
    """Test version import."""
    from invenio_circulation import __version__

    assert __version__


def test_rest_permissions(app, json_headers):
    """Test rest permissions config."""
    with app.test_client() as client:
        url = url_for("invenio_records_rest.loanid_list")
        res = client.post(
            url, data=json.dumps({}), headers=json_headers
        )
        assert res.status_code == 201



# -*- coding: utf-8 -*-
#
# Copyright (C) 2018 CERN.
# Copyright (C) 2018 RERO.
#
# Invenio-Circulation is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Circulation search API."""

from elasticsearch_dsl import VERSION as ES_VERSION
from invenio_search.api import RecordsSearch

from ..proxies import current_circulation


class LoansSearch(RecordsSearch):
    """RecordsSearch for borrowed documents."""

    class Meta:
        """Search only on loans index."""

        index = "loans"
        doc_types = None

    def exclude(self, *args, **kwargs):
        """Add method `exclude` to old elastic search versions."""
        if ES_VERSION[0] == 2:
            from elasticsearch_dsl.query import Bool, Q
            return self.query(Bool(filter=[~Q(*args, **kwargs)]))
        else:
            return super(LoansSearch, self).exclude(*args, **kwargs)


def search_by_pid(item_pid=None, document_pid=None, filter_states=None,
                  exclude_states=None):
    """Retrieve loans attached to the given item or document."""
    search = current_circulation.loan_search

    if document_pid:
        search = search.filter("term", document_pid=document_pid)
    elif item_pid:
        search = search.filter("term", item_pid=item_pid)
    else:
        raise ValueError("Must specify item_pid or document_pid param")

    if filter_states:
        search = search.filter("terms", state=filter_states)
    elif exclude_states:
        search = search.exclude("terms", state=exclude_states)

    return search


def search_by_patron_item(patron_pid, item_pid, filter_states=None):
    """Retrieve loans for patron given an item."""
    search = current_circulation.loan_search
    search = search \
        .filter("term", patron_pid=patron_pid) \
        .filter("term", item_pid=item_pid)

    if filter_states:
        search = search.filter("terms", state=filter_states)

    return search


def search_by_patron_pid(patron_pid):
    """Retrieve loans of a patron."""
    search = current_circulation.loan_search
    search = search.filter("term", patron_pid=patron_pid)
    return search



# -*- coding: utf-8 -*-
#
# Copyright (C) 2018 CERN.
# Copyright (C) 2018 RERO.
#
# Invenio-Circulation is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Invenio Circulation custom transitions."""

from datetime import timedelta

from flask import current_app
from invenio_db import db

from ..api import get_available_item_by_doc_pid, get_document_by_item_pid, \
    get_pending_loans_by_doc_pid
from ..errors import TransitionConditionsFailed, TransitionConstraintsViolation
from ..transitions.base import Transition
from ..transitions.conditions import is_same_location
from ..utils import parse_date


def _ensure_valid_loan_duration(loan):
    """Validate start and end dates for a loan."""
    loan.setdefault('start_date', loan['transaction_date'])

    if not loan.get('end_date'):
        get_loan_duration = current_app.config['CIRCULATION_POLICIES'][
            'checkout']['duration_default']
        number_of_days = get_loan_duration(loan)
        loan['end_date'] = loan['start_date'] + timedelta(days=number_of_days)

    is_duration_valid = current_app.config['CIRCULATION_POLICIES']['checkout'][
        'duration_validate']
    if not is_duration_valid(loan):
        msg = 'The loan duration from `{0}` to `{1}` is not valid'.format(
            loan['start_date'],
            loan['end_date']
        )
        raise TransitionConstraintsViolation(msg=msg)


def _ensure_item_attached_to_loan(loan):
    """Validate that an item is attached to a loan."""
    if not loan.get('item_pid'):
        raise TransitionConditionsFailed(
            msg='No item found attached in loan with pid {0}.'
                .format(str(loan.id)))


def _update_document_pending_request_for_item(item_pid):
    """."""
    document_pid = get_document_by_item_pid(item_pid)
    for loan in get_pending_loans_by_doc_pid(document_pid):
        loan['item_pid'] = item_pid
        loan.commit()
        db.session.commit()
        # TODO: index loan again?


def _ensure_valid_extension(loan):
    """Validate end dates for a extended loan."""
    get_extension_max_count = current_app.config['CIRCULATION_POLICIES'][
        'extension']['max_count']
    extension_max_count = get_extension_max_count(loan)

    extension_count = loan.get('extension_count', 0)
    extension_count += 1
    if extension_count > extension_max_count + 1:
        msg = 'Max extension count reached `{0}`'.format(extension_max_count)
        raise TransitionConstraintsViolation(msg=msg)

    loan['extension_count'] = extension_count

    get_extension_duration = current_app.config['CIRCULATION_POLICIES'][
        'extension']['duration_default']
    number_of_days = get_extension_duration(loan)
    get_extension_from_end_date = current_app.config[
        'CIRCULATION_POLICIES']['extension']['from_end_date']

    end_date = parse_date(loan['end_date'])
    if not get_extension_from_end_date:
        end_date = loan.get('transaction_date')

    end_date += timedelta(days=number_of_days)
    loan['end_date'] = end_date.isoformat()


class ToItemOnLoan(Transition):
    """Action to checkout."""

    def before(self, loan, **kwargs):
        """Validate checkout action."""
        super(ToItemOnLoan, self).before(loan, **kwargs)

        self.ensure_item_is_available(loan)

        if loan.get('start_date'):
            loan['start_date'] = parse_date(loan['start_date'])
        if loan.get('end_date'):
            loan['end_date'] = parse_date(loan['end_date'])

        _ensure_valid_loan_duration(loan)

    def after(self, loan):
        """Convert dates to string before saving loan."""
        loan['start_date'] = loan['start_date'].isoformat()
        loan['end_date'] = loan['end_date'].isoformat()
        super(ToItemOnLoan, self).after(loan)


class ItemAtDeskToItemOnLoan(ToItemOnLoan):
    """Check-out action to perform a loan when item ready at desk."""

    def before(self, loan, **kwargs):
        """Validate checkout action."""
        super(ToItemOnLoan, self).before(loan, **kwargs)

        if loan.get('start_date'):
            loan['start_date'] = parse_date(loan['start_date'])

        if loan.get('end_date'):
            loan['end_date'] = parse_date(loan['end_date'])

        _ensure_valid_loan_duration(loan)


class CreatedToPending(Transition):
    """Action to request to loan an item."""

    def check_request_on_document(f):
        """Decorate to check if the request is on document."""
        def inner(self, loan, **kwargs):
            document_pid = kwargs.get('document_pid')
            if document_pid and not kwargs.get('item_pid'):
                available_item_pid = get_available_item_by_doc_pid(
                    document_pid
                )
                if available_item_pid:
                    kwargs['item_pid'] = available_item_pid
            return f(self, loan, **kwargs)
        return inner

    @check_request_on_document
    def before(self, loan, **kwargs):
        """Set a default pickup location if not passed as param."""
        super(CreatedToPending, self).before(loan, **kwargs)

        # set pickup location to item location if not passed as default
        if not loan.get('pickup_location_pid'):
            item_location_pid = current_app.config[
                'CIRCULATION_ITEM_LOCATION_RETRIEVER'](loan['item_pid'])
            loan['pickup_location_pid'] = item_location_pid


class PendingToItemAtDesk(Transition):
    """Validate pending request to prepare the item at desk of its location."""

    def before(self, loan, **kwargs):
        """Validate if the item is for this location or should transit."""
        super(PendingToItemAtDesk, self).before(loan, **kwargs)

        # check if a request on document has no item attached
        _ensure_item_attached_to_loan(loan)

        if not is_same_location(loan['item_pid'], loan['pickup_location_pid']):
            msg = 'Invalid transition to {0}: Pickup is not at the same ' \
                  'library.'.format(self.dest)
            raise TransitionConditionsFailed(msg=msg)


class PendingToItemInTransitPickup(Transition):
    """Validate pending request to send the item to the pickup location."""

    def before(self, loan, **kwargs):
        """Validate if the item is for this location or should transit."""
        super(PendingToItemInTransitPickup, self).before(loan, **kwargs)

        # check if a request on document has no item attached
        _ensure_item_attached_to_loan(loan)

        if is_same_location(loan['item_pid'], loan['pickup_location_pid']):
            raise TransitionConditionsFailed(
                msg='Invalid transition to {0}: Pickup is at the same library.'
                    .format(self.dest))


class ItemOnLoanToItemOnLoan(Transition):
    """Extend action to perform a item loan extension."""

    def before(self, loan, **kwargs):
        """Validate extension action."""
        super(ItemOnLoanToItemOnLoan, self).before(loan, **kwargs)
        _ensure_valid_extension(loan)


class ItemOnLoanToItemInTransitHouse(Transition):
    """Check-in action when returning an item not to its belonging location."""

    def before(self, loan, **kwargs):
        """Validate check-in action."""
        super(ItemOnLoanToItemInTransitHouse, self).before(loan, **kwargs)
        if is_same_location(loan['item_pid'],
                            loan['transaction_location_pid']):
            msg = 'Invalid transition to {0}: item should be returned ' \
                  'because already to house.'.format(self.dest)
            raise TransitionConditionsFailed(msg=msg)

        # set end loan date as transaction date when completing loan
        loan['end_date'] = loan['transaction_date']

    def after(self, loan):
        """Convert dates to string before saving loan."""
        loan['end_date'] = loan['end_date'].isoformat()
        super(ItemOnLoanToItemInTransitHouse, self).after(loan)


class ItemOnLoanToItemReturned(Transition):
    """Check-in action when returning an item to its belonging location."""

    def before(self, loan, **kwargs):
        """Validate check-in action."""
        super(ItemOnLoanToItemReturned, self).before(loan, **kwargs)
        if not is_same_location(loan['item_pid'],
                                loan['transaction_location_pid']):
            msg = 'Invalid transition to {0}: item should be in transit to ' \
                  'house.'.format(self.dest)
            raise TransitionConditionsFailed(msg=msg)

        # set end loan date as transaction date when completing loan
        loan['end_date'] = loan['transaction_date']

    def after(self, loan):
        """Convert dates to string before saving loan."""
        loan['end_date'] = loan['end_date'].isoformat()
        super(ItemOnLoanToItemReturned, self).after(loan)
        _update_document_pending_request_for_item(loan['item_pid'])


class ItemInTransitHouseToItemReturned(Transition):
    """Check-in action when returning an item to its belonging location."""

    def after(self, loan):
        """Convert dates to string before saving loan."""
        super(ItemInTransitHouseToItemReturned, self).after(loan)
        _update_document_pending_request_for_item(loan['item_pid'])



# -*- coding: utf-8 -*-
#
# Copyright (C) 2018 CERN.
# Copyright (C) 2018 RERO.
#
# Invenio-Circulation is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Circulation exceptions."""


class CirculationException(Exception):
    """Exceptions raised by circulation module."""

    def __init__(self, msg):
        """Constructor."""
        self.msg = msg


class InvalidState(CirculationException):
    """State not found in circulation configuration."""


class ItemNotAvailable(CirculationException):
    """Exception raised from action on unavailable item."""


class TransitionConditionsFailed(CirculationException):
    """Conditions for the transition failed at loan state."""


class NoValidTransitionAvailable(CirculationException):
    """Exception raised when all transitions conditions failed."""


class LoanActionError(CirculationException):
    """Exception raised when invalid action triggered on a loan."""


class InvalidCirculationPermission(CirculationException):
    """Raised when permissions are not satisfied for transition."""


class TransitionConstraintsViolation(CirculationException):
    """Exception raised when constraints for the transition failed."""


class MultipleLoansOnItemError(CirculationException):
    """Exception raised when more than one loan on an item."""



# -*- coding: utf-8 -*-
#
# Copyright (C) 2018 CERN.
# Copyright (C) 2018 RERO.
#
# Invenio-Circulation is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Circulation API."""

from flask import current_app
from invenio_jsonschemas import current_jsonschemas
from invenio_pidstore.resolver import Resolver
from invenio_records.api import Record

from .errors import CirculationException, MultipleLoansOnItemError
from .pidstore.pids import CIRCULATION_LOAN_PID_TYPE
from .search.api import search_by_patron_item, search_by_pid


class Loan(Record):
    """Loan record class."""

    pid_field = "loan_pid"
    _schema = "loans/loan-v1.0.0.json"

    def __init__(self, data, model=None):
        """."""
        self["state"] = current_app.config["CIRCULATION_LOAN_INITIAL_STATE"]
        super(Loan, self).__init__(data, model)

    @classmethod
    def create(cls, data, id_=None, **kwargs):
        """Create Loan record."""
        data["$schema"] = current_jsonschemas.path_to_url(cls._schema)
        ref_builder = current_app.config.get("CIRCULATION_ITEM_REF_BUILDER")
        item_attached = data.get("item_pid") and data["item_pid"] != ""
        if ref_builder and item_attached:
            data["item"] = current_app.config["CIRCULATION_ITEM_REF_BUILDER"](
                data["item_pid"]
            )
        return super(Loan, cls).create(data, id_=id_, **kwargs)

    @classmethod
    def get_record_by_pid(cls, pid, with_deleted=False):
        """Get ils record by pid value."""
        resolver = Resolver(
            pid_type=CIRCULATION_LOAN_PID_TYPE,
            object_type="rec",
            getter=cls.get_record,
        )
        persistent_identifier, record = resolver.resolve(str(pid))
        return record


def is_item_available(item_pid):
    """Return True if the given item is available for loan, False otherwise."""
    config = current_app.config
    cfg_item_available = config["CIRCULATION_POLICIES"]["checkout"].get(
        "item_available"
    )
    if not cfg_item_available(item_pid):
        return False

    search = search_by_pid(
        item_pid=item_pid,
        filter_states=config.get("CIRCULATION_STATES_LOAN_ACTIVE"),
    )
    return search.execute().hits.total == 0


def get_pending_loans_by_item_pid(item_pid):
    """Return any pending loans for the given item."""
    search = search_by_pid(item_pid=item_pid, filter_states=["PENDING"])
    for result in search.scan():
        yield Loan.get_record_by_pid(result[Loan.pid_field])


def get_pending_loans_by_doc_pid(document_pid):
    """Return any pending loans for the given document."""
    search = search_by_pid(document_pid=document_pid,
                           filter_states=["PENDING"])
    for result in search.scan():
        yield Loan.get_record_by_pid(result[Loan.pid_field])


def get_available_item_by_doc_pid(document_pid):
    """Return an item pid available for this document."""
    for item_pid in get_items_by_doc_pid(document_pid):
        if is_item_available(item_pid):
            return item_pid
    return None


def get_items_by_doc_pid(document_pid):
    """Return a list of item pids for this document."""
    return current_app.config["CIRCULATION_ITEMS_RETRIEVER_FROM_DOCUMENT"](
        document_pid
    )


def get_document_by_item_pid(item_pid):
    """Return the document pid of this item_pid."""
    return current_app.config["CIRCULATION_DOCUMENT_RETRIEVER_FROM_ITEM"](
        item_pid
    )


def get_loan_for_item(item_pid):
    """Return the Loan attached to the given item, if any."""
    loan = None
    if not item_pid:
        return

    search = search_by_pid(
        item_pid=item_pid,
        filter_states=current_app.config["CIRCULATION_STATES_LOAN_ACTIVE"],
    )

    hits = list(search.scan())
    if hits:
        if len(hits) > 1:
            raise MultipleLoansOnItemError(
                "Multiple active loans on item {0}".format(item_pid)
            )

        loan = Loan.get_record_by_pid(hits[0][Loan.pid_field])
    return loan


def patron_has_active_loan_on_item(patron_pid, item_pid):
    """Return True if patron has a pending or active Loan for given item."""
    config = current_app.config
    if not patron_pid or not item_pid:
        raise CirculationException("Patron PID or Item PID not specified")

    states = ["CREATED", "PENDING"] + config["CIRCULATION_STATES_LOAN_ACTIVE"]
    search = search_by_patron_item(
        patron_pid=patron_pid,
        item_pid=item_pid,
        filter_states=states,
    )
    search_result = search.execute()
    return search_result.hits.total > 0



# -*- coding: utf-8 -*-
#
# Copyright (C) 2018 CERN.
# Copyright (C) 2018 RERO.
#
# Invenio-Circulation is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.


"""Test circulation permissions on transitions."""

import pytest
from invenio_records_rest.utils import allow_all, deny_all

from invenio_circulation.errors import InvalidCirculationPermission
from invenio_circulation.transitions.transitions import CreatedToPending


def test_valid_permission(loan_created, params):
    """Test transition with valid permission."""
    transition = CreatedToPending(
        'CREATED', 'PENDING', trigger='next', permission_factory=allow_all
    )
    transition.execute(loan_created, **params)
    assert loan_created['state'] == 'PENDING'


def test_invalid_permission(loan_created, params):
    """Test transition without permission."""
    transition = CreatedToPending(
        'CREATED', 'PENDING', trigger='next', permission_factory=deny_all
    )
    with pytest.raises(InvalidCirculationPermission):
        transition.execute(loan_created, **params)
    assert loan_created['state'] == 'CREATED'



# -*- coding: utf-8 -*-
#
# Copyright (C) 2018 CERN.
# Copyright (C) 2018 RERO.
#
# Invenio-Circulation is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Circulation PIDs."""

CIRCULATION_LOAN_PID_TYPE = 'loanid'
"""Persistent Identifier for Loans."""

CIRCULATION_LOAN_MINTER = 'loanid'
"""Minter PID for Loans."""

CIRCULATION_LOAN_FETCHER = 'loanid'
"""Fetcher PID for Loans."""

_LOANID_CONVERTER = 'pid(loanid,record_class="invenio_circulation.api:Loan")'
"""Loan PID url converter."""



# -*- coding: utf-8 -*-
#
# Copyright (C) 2018 CERN.
# Copyright (C) 2018 RERO.
#
# Invenio-Circulation is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Circulation PID providers."""

from invenio_pidstore.models import PIDStatus
from invenio_pidstore.providers.recordid import RecordIdProvider

from .pids import CIRCULATION_LOAN_PID_TYPE


class CirculationLoanIdProvider(RecordIdProvider):
    """Record identifier provider."""

    pid_type = CIRCULATION_LOAN_PID_TYPE
    """Type of persistent identifier."""

    pid_provider = None
    """Provider name.

    The provider name is not recorded in the PID since the provider does not
    provide any additional features besides creation of record ids.
    """

    default_status = PIDStatus.REGISTERED
    """Record IDs are by default registered immediately."""



# -*- coding: utf-8 -*-
#
# Copyright (C) 2018 CERN.
# Copyright (C) 2018 RERO.
#
# Invenio-Circulation is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Invenio Circulation base transitions."""

from datetime import datetime

from flask import current_app
from invenio_db import db

from ..api import is_item_available
from ..errors import InvalidCirculationPermission, InvalidState, \
    ItemNotAvailable, TransitionConditionsFailed, \
    TransitionConstraintsViolation
from ..proxies import current_circulation
from ..signals import loan_state_changed
from ..utils import parse_date


def ensure_same_item_patron(f):
    """Validate that item and patron PIDs exist and cannot be changed."""
    def inner(self, loan, **kwargs):
        new_patron_pid = kwargs.get('patron_pid')
        new_item_pid = kwargs.get('item_pid')

        if not current_app.config['CIRCULATION_ITEM_EXISTS'](new_item_pid):
            msg = 'Item `{0}` not found in the system'.format(new_item_pid)
            raise TransitionConstraintsViolation(msg=msg)

        if loan.get('item_pid') and new_item_pid != loan['item_pid']:
            msg = 'Loan item is `{0}` but transition is trying to set it to ' \
                  '`{1}`'.format(loan['item_pid'], new_item_pid)
            raise TransitionConstraintsViolation(msg=msg)

        if not current_app.config['CIRCULATION_PATRON_EXISTS'](new_patron_pid):
            msg = 'Patron `{0}` not found in the system'.format(new_patron_pid)
            raise TransitionConstraintsViolation(msg=msg)

        if 'patron_pid' in loan and new_patron_pid != loan['patron_pid']:
            msg = 'Loan patron is `{0}` but transition is trying to set it ' \
                  'to `{1}`'.format(loan['patron_pid'], new_patron_pid)
            raise TransitionConstraintsViolation(msg=msg)

        return f(self, loan, **kwargs)
    return inner


def ensure_required_params(f):
    """Decorate to ensure that all required parameters has been passed."""
    def inner(self, loan, **kwargs):
        missing = [p for p in self.REQUIRED_PARAMS if p not in kwargs]
        if missing:
            msg = 'Required input parameters are missing `[{}]`'\
                .format(missing)
            raise TransitionConstraintsViolation(msg=msg)
        if all(param not in kwargs for param in self.PARTIAL_REQUIRED_PARAMS):
            msg = 'One of the parameters `[{}]` must be passed.'\
                .format(missing)
            raise TransitionConstraintsViolation(msg=msg)
        return f(self, loan, **kwargs)
    return inner


def check_trigger(f):
    """Decorate to check the transition should be manually triggered."""
    def inner(self, loan, **kwargs):
        if kwargs.get('trigger', 'next') != self.trigger:
            msg = 'No param `trigger` with value `{0}`.'.format(self.trigger)
            raise TransitionConditionsFailed(msg=msg)
        return f(self, loan, **kwargs)
    return inner


class Transition(object):
    """A transition object that is triggered on conditions."""

    REQUIRED_PARAMS = [
        'transaction_user_pid',
        'patron_pid',
        'transaction_location_pid',
        'transaction_date'
    ]

    PARTIAL_REQUIRED_PARAMS = [
        'item_pid',
        'document_pid'
    ]

    def __init__(self, src, dest, trigger='next', permission_factory=None,
                 **kwargs):
        """Init transition object."""
        self.src = src
        self.dest = dest
        self.trigger = trigger
        self.permission_factory = permission_factory or current_app.config[
            'CIRCULATION_LOAN_TRANSITIONS_DEFAULT_PERMISSION_FACTORY']
        self.validate_transition_states()

    def ensure_item_is_available(self, loan):
        """Validate that an item is available."""
        if 'item_pid' not in loan:
            msg = 'Item not set for loan'.format(loan['loan_pid'])
            raise TransitionConstraintsViolation(msg=msg)

        if not is_item_available(loan['item_pid']):
            msg = 'Invalid transition to {0}: Item {1} is unavailable.'\
                .format(self.dest, loan['item_pid'])
            raise ItemNotAvailable(msg)

    def validate_transition_states(self):
        """Ensure that source and destination states are valid."""
        states = current_app.config['CIRCULATION_LOAN_TRANSITIONS'].keys()
        if not all([self.src in states, self.dest in states]):
            msg = 'Source state `{0}` or destination state `{1}` not in [{2}]'\
                .format(self.src, self.dest, states)
            raise InvalidState(msg=msg)

    @check_trigger
    @ensure_required_params
    @ensure_same_item_patron
    def before(self, loan, **kwargs):
        """Validate input, evaluate conditions and raise if failed."""
        if self.permission_factory and not self.permission_factory(loan).can():
            msg = 'Invalid circulation permission'
            raise InvalidCirculationPermission(msg=msg)

        kwargs.setdefault('transaction_date', datetime.now())
        kwargs['transaction_date'] = parse_date(kwargs['transaction_date'])
        loan.update(kwargs)

    def execute(self, loan, **kwargs):
        """Execute before actions, transition and after actions."""
        self.before(loan, **kwargs)
        loan['state'] = self.dest
        self.after(loan)

    def after(self, loan):
        """Commit record and index."""
        loan['transaction_date'] = loan['transaction_date'].isoformat()
        loan.commit()
        db.session.commit()
        current_circulation.loan_indexer.index(loan)

        loan_state_changed.send(self, loan=loan)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2018 CERN.
# Copyright (C) 2018 RERO.
#
# Invenio-Circulation is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Tests for loan states."""

import json

from flask import url_for
from invenio_indexer.api import RecordIndexer
from invenio_search import current_search

from .helpers import create_loan


def _get(app, json_headers, pid_value):
    """Perform API GET with the given param."""
    with app.test_client() as client:
        url = url_for('invenio_circulation_loan_for_item.loan_resource',
                      pid_value=pid_value)
        res = client.get(url, headers=json_headers)

    payload = None
    if res.status_code == 200:
        payload = json.loads(res.data.decode('utf-8'))
    return res, payload


def test_rest_circulation_item_loan_pending(app, json_headers, indexed_loans):
    """Test API GET call to check item's pending loan."""
    pending_item_pid = "item_pending_1"
    res, payload = _get(app, json_headers, pending_item_pid)
    assert res.status_code == 200
    assert payload == {}


def test_rest_circulation_item_no_loan(app, json_headers, indexed_loans):
    """Test API GET call to check circulation when no loan on the item."""
    not_loaned_item_pid = "item_not_loaned"
    res, payload = _get(app, json_headers, not_loaned_item_pid)
    assert res.status_code == 200
    assert payload == {}


def test_rest_circulation_item_on_loan(app, json_headers, indexed_loans):
    """Test API GET call to check circulation item's loan."""
    multiple_loans_pid = "item_multiple_pending_on_loan_7"
    res, payload = _get(app, json_headers, multiple_loans_pid)
    assert res.status_code == 200
    loan = payload['metadata']
    assert loan['state'] == 'ITEM_ON_LOAN'
    assert loan['item_pid'] == multiple_loans_pid


def test_rest_circulation_item_not_found(app, json_headers, indexed_loans):
    """Test API GET call to check circulation item not found."""
    res, payload = _get(app, json_headers, "")
    assert res.status_code == 404


def test_multiple_active_loans(app, db, json_headers, indexed_loans):
    """Test return error when more than one loan on given item."""
    multiple_loans_pid = "item_multiple_pending_on_loan_7"

    test_loan_data = {
        "item_pid": "item_multiple_pending_on_loan_7",
        "patron_pid": "2",
        "state": "ITEM_ON_LOAN",
        "transaction_date": "2018-06-26",
        "transaction_location_pid": "loc_pid",
        "transaction_user_pid": "user_pid",
        "start_date": "2018-07-24",
        "end_date": "2018-08-23"
    }

    pid, loan = create_loan(test_loan_data)
    db.session.commit()
    RecordIndexer().index(loan)
    current_search.flush_and_refresh(index="loans")

    res, payload = _get(app, json_headers, multiple_loans_pid)
    assert res.status_code == 500



# -*- coding: utf-8 -*-
#
# Copyright (C) 2018 CERN.
# Copyright (C) 2018 RERO.
#
# Invenio-Circulation is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Circulation fetchers."""

from invenio_pidstore.fetchers import FetchedPID

from ..api import Loan
from .pids import CIRCULATION_LOAN_PID_TYPE


def loan_pid_fetcher(record_uuid, data):
    """Fetch PID from loan record."""
    return FetchedPID(
        provider=None,
        pid_type=CIRCULATION_LOAN_PID_TYPE,
        pid_value=str(data[Loan.pid_field])
    )



# -*- coding: utf-8 -*-
#
# Copyright (C) 2018 CERN.
#
# invenio-circulation is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Circulation permissions."""

from __future__ import absolute_import, print_function

from functools import wraps

from flask import abort, current_app
from flask_login import current_user
from invenio_access import action_factory
from invenio_access.permissions import Permission
from invenio_records_rest.utils import allow_all

loan_access = action_factory('loan-read-access')


def check_permission(permission):
    """Abort if permission is not allowed.

    :param permission: The permission to check.
    """
    if permission is not None and not permission.can():
        if current_user.is_authenticated:
            abort(403, 'You do not have a permission for this action')
        abort(401)


def has_read_loan_permission(*args, **kwargs):
    """Return permission to allow user to access loan."""
    return Permission(loan_access)


def loan_reader(*args, **kwargs):
    """Return an object which by default allows to read the load object."""
    def can(self):
        """Return True if user can read the loan record."""
        return has_read_loan_permission(*args, **kwargs).can()

    return type('Allow', (), {'can': can})()


def login_required(*args, **kwargs):
    """Return an object that evaluates if the current user is authenticated."""
    def can(self):
        """Return True if user is authenticated."""
        return current_user.is_authenticated

    return type('LoginRequired', (), {'can': can})()


def views_permissions_factory(action):
    """Circulation views permissions factory."""
    if action == 'loan-read-access':
        return allow_all()
    elif action == 'loan-actions':
        return allow_all()


def need_permissions(action):
    """View decorator to check permissions for the given action or abort.

    :param action: The action needed.
    """
    def decorator_builder(f):
        @wraps(f)
        def decorate(*args, **kwargs):
            check_permission(
                current_app
                .config['CIRCULATION_VIEWS_PERMISSIONS_FACTORY'](action)
            )
            return f(*args, **kwargs)
        return decorate
    return decorator_builder



# -*- coding: utf-8 -*-
#
# Copyright (C) 2018 CERN.
# Copyright (C) 2018 RERO.
#
# Invenio-Circulation is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Tests for loan states."""

from datetime import timedelta

import mock
import pytest
from flask import current_app
from flask_security import login_user

from invenio_circulation.api import Loan, is_item_available
from invenio_circulation.errors import ItemNotAvailable, \
    NoValidTransitionAvailable, TransitionConstraintsViolation
from invenio_circulation.proxies import current_circulation
from invenio_circulation.utils import parse_date

from .helpers import SwappedConfig, SwappedNestedConfig


@mock.patch(
    "invenio_circulation.transitions.transitions"
    ".get_pending_loans_by_doc_pid"
)
def test_loan_checkout_checkin(
    mock_pending_loans_for_document,
    loan_created,
    db,
    params,
    mock_is_item_available,
):
    """Test loan checkout and checkin actions."""
    mock_pending_loans_for_document.return_value = []
    assert loan_created["state"] == "CREATED"

    loan = current_circulation.circulation.trigger(
        loan_created, **dict(params, trigger="checkout")
    )
    db.session.commit()
    assert loan["state"] == "ITEM_ON_LOAN"

    # set same transaction location to avoid "in transit"
    same_location = params["transaction_location_pid"]
    with SwappedConfig(
        "CIRCULATION_ITEM_LOCATION_RETRIEVER", lambda x: same_location
    ):
        loan = current_circulation.circulation.trigger(loan, **dict(params))
        db.session.commit()
        assert loan["state"] == "ITEM_RETURNED"


def test_loan_request(loan_created, db, params):
    """Test loan request action."""
    assert loan_created["state"] == "CREATED"

    loan = current_circulation.circulation.trigger(
        loan_created,
        **dict(
            params,
            trigger="request",
            pickup_location_pid="pickup_location_pid",
        )
    )
    db.session.commit()
    assert loan["state"] == "PENDING"


def test_loan_extend(loan_created, db, params, mock_is_item_available):
    """Test loan extend action."""
    def get_max_count_1(loan):
        return 1

    loan = current_circulation.circulation.trigger(
        loan_created, **dict(params, trigger="checkout")
    )
    db.session.commit()
    end_date = parse_date(loan["end_date"])

    loan = current_circulation.circulation.trigger(
        loan, **dict(params, trigger="extend")
    )
    db.session.commit()
    new_end_date = parse_date(loan["end_date"])
    assert new_end_date == end_date + timedelta(days=30)
    assert loan["extension_count"] == 1
    loan = current_circulation.circulation.trigger(
        loan, **dict(params, trigger="extend")
    )
    db.session.commit()

    # test to manny extensions
    current_app.config["CIRCULATION_POLICIES"]["extension"][
        "max_count"
    ] = get_max_count_1
    with pytest.raises(TransitionConstraintsViolation):
        loan = current_circulation.circulation.trigger(
            loan, **dict(params, trigger="extend")
        )


def test_loan_extend_from_enddate(
    loan_created, db, params, mock_is_item_available
):
    """Test loan extend action from transaction date."""
    loan = current_circulation.circulation.trigger(
        loan_created, **dict(params, trigger="checkout")
    )
    db.session.commit()
    extension_date = parse_date(loan.get("transaction_date"))
    current_app.config["CIRCULATION_POLICIES"]["extension"][
        "from_end_date"
    ] = False

    loan = current_circulation.circulation.trigger(
        loan, **dict(params, trigger="extend")
    )
    db.session.commit()
    new_end_date = parse_date(loan["end_date"])
    assert new_end_date == extension_date + timedelta(days=30)
    assert loan["extension_count"] == 1


def test_cancel_action(loan_created, db, params, mock_is_item_available):
    """Test should pass when calling `cancel` from `ITEM_ON_LOAN`."""
    loan = current_circulation.circulation.trigger(
        loan_created, **dict(params, trigger="checkout")
    )
    db.session.commit()

    current_circulation.circulation.trigger(
        loan_created, **dict(params, trigger="cancel")
    )
    assert loan["state"] == "CANCELLED"


def test_cancel_fail(loan_created, params):
    """Test should fail when calling `cancel` from `CREATED`."""
    with pytest.raises(NoValidTransitionAvailable):
        current_circulation.circulation.trigger(
            loan_created, **dict(params, trigger="cancel")
        )


def test_validate_item_in_transit_for_pickup(loan_created, db, params):
    """."""
    loan = current_circulation.circulation.trigger(
        loan_created,
        **dict(
            params,
            trigger="request",
            pickup_location_pid="pickup_location_pid",
        )
    )
    db.session.commit()
    assert loan["state"] == "PENDING"

    with SwappedConfig(
        "CIRCULATION_ITEM_LOCATION_RETRIEVER",
        lambda x: "external_location_pid",
    ):
        loan = current_circulation.circulation.trigger(loan, **dict(params))
        assert loan["state"] == "ITEM_IN_TRANSIT_FOR_PICKUP"


def test_validate_item_at_desk(loan_created, db, params):
    """."""
    loan = current_circulation.circulation.trigger(
        loan_created,
        **dict(
            params,
            trigger="request",
            pickup_location_pid="pickup_location_pid",
        )
    )
    db.session.commit()
    assert loan["state"] == "PENDING"

    with SwappedConfig(
        "CIRCULATION_ITEM_LOCATION_RETRIEVER", lambda x: "pickup_location_pid"
    ):
        loan = current_circulation.circulation.trigger(
            loan_created, **dict(params)
        )
        assert loan["state"] == "ITEM_AT_DESK"


def test_checkout_start_is_transaction_date(
    loan_created, db, params, mock_is_item_available
):
    """Test checkout start date to transaction date when not set."""
    number_of_days = 10

    with SwappedNestedConfig(
        ["CIRCULATION_POLICIES", "checkout", "duration_default"],
        lambda x: number_of_days,
    ):
        loan = current_circulation.circulation.trigger(
            loan_created, **dict(params, trigger="checkout")
        )
        db.session.commit()

        assert loan["state"] == "ITEM_ON_LOAN"
        assert loan["start_date"] == loan["transaction_date"]
        start_date = parse_date(loan["start_date"])
        end_date = start_date + timedelta(number_of_days)
        assert loan["end_date"] == end_date.isoformat()


def test_checkout_with_input_start_end_dates(
    loan_created, db, params, mock_is_item_available
):
    """Test checkout start and end dates are set as input."""
    start_date = "2018-02-01T09:30:00+02:00"
    end_date = "2018-02-10T09:30:00+02:00"
    loan = current_circulation.circulation.trigger(
        loan_created,
        **dict(
            params,
            start_date=start_date,
            end_date=end_date,
            trigger="checkout",
        )
    )
    db.session.commit()
    assert loan["state"] == "ITEM_ON_LOAN"
    assert loan["start_date"] == start_date
    assert loan["end_date"] == end_date


def test_checkout_fails_when_wrong_dates(
    loan_created, params, mock_is_item_available
):
    """Test checkout fails when wrong input dates."""
    with pytest.raises(ValueError):
        current_circulation.circulation.trigger(
            loan_created,
            **dict(
                params,
                start_date="2018-xx",
                end_date="2018-xx",
                trigger="checkout",
            )
        )


def test_checkout_fails_when_duration_invalid(
    loan_created, params, mock_is_item_available
):
    """Test checkout fails when wrong max duration."""
    with pytest.raises(TransitionConstraintsViolation):
        with SwappedNestedConfig(
            ["CIRCULATION_POLICIES", "checkout", "duration_validate"],
            lambda x: False,
        ):
            current_circulation.circulation.trigger(
                loan_created,
                **dict(
                    params,
                    start_date="2018-02-01T09:30:00+02:00",
                    end_date="2018-04-10T09:30:00+02:00",
                    trigger="checkout",
                )
            )


@mock.patch(
    "invenio_circulation.transitions.transitions"
    ".get_pending_loans_by_doc_pid"
)
def test_checkin_end_date_is_transaction_date(
    mock_pending_loans_for_document,
    loan_created,
    db,
    params,
    mock_is_item_available,
):
    """Test date the checkin date is the transaction date."""
    mock_pending_loans_for_document.return_value = []
    loan = current_circulation.circulation.trigger(
        loan_created,
        **dict(
            params,
            start_date="2018-02-01T09:30:00+02:00",
            end_date="2018-02-10T09:30:00+02:00",
            trigger="checkout",
        )
    )
    db.session.commit()
    assert loan["state"] == "ITEM_ON_LOAN"

    same_location = params["transaction_location_pid"]
    with SwappedConfig(
        "CIRCULATION_ITEM_LOCATION_RETRIEVER", lambda x: same_location
    ):
        params["transaction_date"] = "2018-03-11T19:15:00+02:00"
        loan = current_circulation.circulation.trigger(loan, **dict(params))
        assert loan["state"] == "ITEM_RETURNED"
        assert loan["end_date"] == params["transaction_date"]


def test_item_availability(indexed_loans):
    """Test item_availability with various conditions."""
    assert is_item_available(item_pid="item_pending_1")
    assert not is_item_available(item_pid="item_on_loan_2")
    assert is_item_available(item_pid="item_returned_3")
    assert not is_item_available(item_pid="item_in_transit_4")
    assert not is_item_available(item_pid="item_at_desk_5")
    assert not is_item_available(item_pid="item_pending_on_loan_6")
    assert is_item_available(item_pid="item_returned_6")
    assert is_item_available(item_pid="no_loan")


def test_checkout_on_unavailable_item(
    loan_created, db, params, mock_is_item_available
):
    """Test checkout fails on unvailable item."""
    mock_is_item_available.return_value = False

    with pytest.raises(ItemNotAvailable):
        loan = current_circulation.circulation.trigger(
            loan_created, **dict(params, trigger="checkout")
        )

        loan_created["state"] = "ITEM_AT_DESK"

        loan = current_circulation.circulation.trigger(
            loan_created, **dict(params)
        )


def _loan_steps_created_to_on_loan(loan_created, params):
    """Go step by step through all the transitions"""

    # loan created
    loan = current_circulation.circulation.trigger(
        loan_created, **dict(params, trigger="request")
    )

    # loan pending
    loan = current_circulation.circulation.trigger(
        loan_created, **dict(params, trigger="next")
    )

    # item at desk
    loan = current_circulation.circulation.trigger(
        loan_created, **dict(params)
    )
    assert loan_created["state"] == "ITEM_ON_LOAN"


def test_checkout_item(loan_created, db, params,
                       users, app):
    """Test standard checkout procedure."""
    login_user(users['user'])
    # set specific loan owner
    params["patron_pid"] = "3"
    _loan_steps_created_to_on_loan(loan_created, params)


def test_checkout_item_unavailable_steps(loan_created, db, params,
                                         users, app):
    """Test checkout attempt on unavailable item."""
    user = users['manager']
    login_user(user)
    with pytest.raises(NoValidTransitionAvailable):
        # loan created
        loan = current_circulation.circulation.trigger(
            loan_created, **dict(params, trigger="request")
        )

        # loan pending
        loan = current_circulation.circulation.trigger(
            loan_created, **dict(params, trigger="next")
        )

        loan_created["state"] = "ITEM_ON_LOAN"
        loan = current_circulation.circulation.trigger(
            loan_created, **dict(params)
        )

        # trying to checkout item already on loan
        loan = current_circulation.circulation.trigger(
            loan_created, **dict(params, trigger="checkout")
        )


@mock.patch("invenio_circulation.api.is_item_available")
def test_request_on_document_with_available_items(
    mock_available_item, loan_created, db, params
):
    """Test loan request action."""
    mock_available_item.return_value = True
    with SwappedConfig(
        "CIRCULATION_ITEMS_RETRIEVER_FROM_DOCUMENT", lambda x: ["item_pid"]
    ):
        loan = current_circulation.circulation.trigger(
            loan_created,
            **dict(
                params,
                trigger="request",
                item_pid=None,
                document_pid="document_pid",
                pickup_location_pid="pickup_location_pid",
            )
        )
        db.session.commit()
        assert loan["state"] == "PENDING"
        assert loan["item_pid"] == "item_pid"
        assert loan["document_pid"] == "document_pid"


@mock.patch("invenio_circulation.api.is_item_available")
def test_request_on_document_with_unavailable_items(
    mock_available_item, loan_created, db, params
):
    """Test loan request action."""
    mock_available_item.return_value = False
    with SwappedConfig(
        "CIRCULATION_ITEMS_RETRIEVER_FROM_DOCUMENT", lambda x: ["item_pid"]
    ):
        # remove item_pid
        params.pop("item_pid")
        loan = current_circulation.circulation.trigger(
            loan_created,
            **dict(
                params,
                trigger="request",
                document_pid="document_pid",
                pickup_location_pid="pickup_location_pid",
            )
        )
        db.session.commit()
        assert loan["state"] == "PENDING"
        assert "item_pid" not in loan
        assert loan["document_pid"] == "document_pid"


@mock.patch(
    "invenio_circulation.transitions.transitions"
    ".get_pending_loans_by_doc_pid"
)
@mock.patch("invenio_circulation.api.is_item_available")
def test_document_requests_on_item_returned(
    mock_available_item,
    mock_pending_loans_for_document,
    mock_is_item_available,
    loan_created,
    db,
    params,
):
    """Test loan request action."""

    # return item is not available
    mock_available_item.return_value = False

    with SwappedConfig(
        "CIRCULATION_DOCUMENT_RETRIEVER_FROM_ITEM", lambda x: "document_pid"
    ):
        same_location = params["transaction_location_pid"]
        with SwappedConfig(
            "CIRCULATION_ITEM_LOCATION_RETRIEVER", lambda x: same_location
        ):
            # start a loan on item with pid 'item_pid'
            new_loan = current_circulation.circulation.trigger(
                loan_created,
                **dict(
                    params,
                    trigger="checkout",
                    item_pid="item_pid",
                    pickup_location_pid="pickup_location_pid",
                )
            )
            db.session.commit()
            assert new_loan["state"] == "ITEM_ON_LOAN"

            # create a new loan request on document_pid without items available
            new_loan_created = Loan.create({
                Loan.pid_field: "2"
            })
            # remove item_pid
            params.pop("item_pid")
            pending_loan = current_circulation.circulation.trigger(
                new_loan_created,
                **dict(
                    params,
                    trigger="request",
                    document_pid="document_pid",
                    pickup_location_pid="pickup_location_pid",
                )
            )
            db.session.commit()
            assert pending_loan["state"] == "PENDING"
            # no item available found. Request is created with no item attached
            assert "item_pid" not in pending_loan
            assert pending_loan["document_pid"] == "document_pid"

            # resolve pending document requests to `document_pid`
            mock_pending_loans_for_document.return_value = [pending_loan]

            returned_loan = current_circulation.circulation.trigger(
                new_loan,
                **dict(
                    params,
                    item_pid="item_pid",
                    pickup_location_pid="pickup_location_pid",
                )
            )
            db.session.commit()
            assert returned_loan["state"] == "ITEM_RETURNED"

            # item `item_pid` has been attached to pending loan request on
            # `document_pid` automatically
            assert pending_loan["state"] == "PENDING"
            assert pending_loan["item_pid"] == "item_pid"
            assert pending_loan["document_pid"] == "document_pid"



# -*- coding: utf-8 -*-
#
# Copyright (C) 2018 CERN.
# Copyright (C) 2018 RERO.
#
# Invenio-Circulation is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Test API permissions"""

import json

from flask import url_for
from invenio_accounts.models import User
from invenio_accounts.testutils import login_user_via_session


def _get(app, json_headers, pid_value, user_email=None):
    """Perform API GET with the given param."""
    with app.test_client() as client:
        if user_email:
            login_user_via_session(client, email=user_email)

        url = url_for('invenio_circulation_loan_for_item.loan_resource',
                      pid_value=pid_value)
        res = client.get(url, headers=json_headers)

    payload = None
    if res.status_code == 200:
        payload = json.loads(res.data.decode('utf-8'))
    return res, payload


def test_default_permission_factory(app, json_headers, indexed_loans):
    """Test default permission factory."""
    multiple_loans_pid = "item_multiple_pending_on_loan_7"
    res, _ = _get(app, json_headers, multiple_loans_pid)
    assert res.status_code == 200


def test_loan_access_anonymous(app, json_headers, indexed_loans, users):
    """Test access for anonymous user to items loan endpoint."""
    multiple_loans_pid = "item_multiple_pending_on_loan_7"
    res, _ = _get(app, json_headers, multiple_loans_pid)
    assert res.status_code == 401


def test_logged_user_no_access(app, json_headers, indexed_loans, users):
    """Test access for user with no right to see the loan."""
    user = users['user']
    user_email = User.query.get(user.id).email

    multiple_loans_pid = "item_multiple_pending_on_loan_7"
    res, _ = _get(app, json_headers, multiple_loans_pid, user_email)
    assert res.status_code == 403


def test_logged_user_access_granted(app, json_headers, indexed_loans, users):
    """Test access for user with a right to see the loan."""
    user = users['manager']
    user_email = User.query.get(user.id).email

    multiple_loans_pid = "item_multiple_pending_on_loan_7"
    res, payload = _get(app, json_headers, multiple_loans_pid, user_email)
    assert res.status_code == 200
    assert payload['metadata']['item_pid'] == multiple_loans_pid
